#include <stdbool.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "rvl.h"

#include "detail/rvl_compress_p.h"
#include "detail/rvl_p.h"
#include "detail/rvl_text_p.h"
#include "util/logger.h"

static void rvl_write_chunk_header (RVL *self, u32 code, u32 size);
static void rvl_write_chunk_payload (RVL *self, const BYTE *payload, u32 size);
static void rvl_write_chunk_end (RVL *self);

static void rvl_handle_VFMT_chunk (RVL *self);
static void rvl_handle_GRID_chunk (RVL *self);
static void rvl_handle_DATA_chunk (RVL *self);
static void rvl_handle_TEXT_chunk (RVL *self);
static void rvl_handle_VEND_chunk (RVL *self);

static void rvl_write_file_sig (RVL *self);
static void rvl_fwrite (RVL *self, const BYTE *data, u32 size);

static void check_grid (RVL *self);
static void check_data (RVL *self);

void
rvl_write_rvl (RVL *self)
{
  check_data (self);
  check_grid (self);

  rvl_write_file_sig (self);

  // Required chunks
  rvl_handle_VFMT_chunk (self);
  rvl_handle_GRID_chunk (self);
  rvl_handle_DATA_chunk (self);

  if (self->text != NULL)
    {
      rvl_handle_TEXT_chunk (self);
    }

  rvl_handle_VEND_chunk (self);
}

void
rvl_handle_VFMT_chunk (RVL *self)
{
  u32 size = 18;
  u8 *wbuf = calloc (1, size);

  RVLPrimitive primitive = self->primitive;
  RVLEndian    endian    = self->endian;
  RVLCompress  compress  = self->compress;

  memcpy (&wbuf[0], self->version, 2);
  memcpy (&wbuf[2], &self->resolution[0], 12);
  memcpy (&wbuf[14], &primitive, 2);
  memcpy (&wbuf[16], &endian, 1);
  memcpy (&wbuf[17], &compress, 1);

  rvl_write_chunk_header (self, RVL_CHUNK_CODE_VFMT, size);
  rvl_write_chunk_payload (self, wbuf, size);
  rvl_write_chunk_end (self);

  free (wbuf);
}

void
rvl_handle_GRID_chunk (RVL *self)
{
  u32   offset   = 14;
  u32   wbufSize = offset + self->grid.dimBufSz;
  BYTE *wbuf     = (BYTE *)malloc (wbufSize);

  RVLGridType type = self->grid.type;
  RVLGridUnit unit = self->grid.unit;

  // Grid info
  wbuf[0] = type;
  wbuf[1] = unit;
  memcpy (&wbuf[2], self->grid.position, 12);

  u32 szdx = self->grid.ndx * sizeof (f32);
  u32 szdy = self->grid.ndy * sizeof (f32);
  u32 szdz = self->grid.ndz * sizeof (f32);

  memcpy (&wbuf[offset], self->grid.dx, szdx);
  offset += szdx;
  memcpy (&wbuf[offset], self->grid.dy, szdy);
  offset += szdy;
  memcpy (&wbuf[offset], self->grid.dz, szdz);

  rvl_write_chunk_header (self, RVL_CHUNK_CODE_GRID, wbufSize);
  rvl_write_chunk_payload (self, wbuf, wbufSize);
  rvl_write_chunk_end (self);

  free (wbuf);
}

void
rvl_handle_DATA_chunk (RVL *self)
{
  u32   wbufSize = self->data.size;
  BYTE *wbuf     = (BYTE *)malloc (wbufSize);

  if (self->compress == RVL_COMPRESSION_LZMA2)
    {
      rvl_compress_lzma (self, &wbuf, &wbufSize);
    }
  else if (self->compress == RVL_COMPRESSION_LZ4)
    {
      rvl_compress_lz4 (self, &wbuf, &wbufSize);
    }

  rvl_write_chunk_header (self, RVL_CHUNK_CODE_DATA, wbufSize);
  rvl_write_chunk_payload (self, (const BYTE *)wbuf, wbufSize);
  rvl_write_chunk_end (self);

  free (wbuf);
}

// Strip off the null terminator at the end of the value string.
void
rvl_handle_TEXT_chunk (RVL *self)
{
  const RVLText *cur = self->text;
  while (cur != NULL)
    {
      const RVLText *const text      = cur;
      const u32            valueSize = strlen (text->value);

      rvl_write_chunk_header (self, RVL_CHUNK_CODE_TEXT, valueSize + 1);
      rvl_write_chunk_payload (self, (const BYTE *)&cur->tag, 1);

      if (valueSize != 0)
        {
          rvl_write_chunk_payload (self, (const BYTE *)text->value, valueSize);
        }

      rvl_write_chunk_end (self);

      cur = cur->next;
    }
}

void
rvl_handle_VEND_chunk (RVL *self)
{
  rvl_write_chunk_header (self, RVL_CHUNK_CODE_VEND, 0);
  rvl_write_chunk_payload (self, NULL, 0);
  rvl_write_chunk_end (self);
}

void
rvl_write_chunk_header (RVL *self, u32 code, u32 size)
{
  u32 buf[2];

  buf[0] = size;
  buf[1] = code;

  rvl_reset_crc32 (self);
  rvl_calculate_crc32 (self, (BYTE *)&code, 4);
  rvl_fwrite (self, (const BYTE *)buf, sizeof (buf));
}

void
rvl_write_chunk_payload (RVL *self, const BYTE *payload, u32 size)
{
  if (payload != NULL && size > 0)
    {
      rvl_calculate_crc32 (self, payload, size);
      rvl_fwrite (self, payload, size);
    }
}

void
rvl_write_chunk_end (RVL *self)
{
  BYTE buf[4];
  rvl_save_u32le (buf, self->crc);
  rvl_fwrite (self, buf, sizeof (u32));
  return;
}

void
rvl_write_file_sig (RVL *self)
{
  rvl_fwrite (self, RVL_FILE_SIG, RVL_FILE_SIG_SIZE);
}

void
rvl_fwrite (RVL *self, const BYTE *data, u32 size)
{
  if (self->writeFn == NULL)
    {
      log_fatal ("Call to NULL write function. Please check if "
                     "the RVL instance is a writer.");
      exit (EXIT_FAILURE);
    }

  self->writeFn (self, data, size);
}

void
rvl_fwrite_default (RVL *self, const BYTE *data, u32 size)
{
  const size_t count = fwrite (data, 1, size, self->io);

  if (count != size)
    {
      log_fatal ("Failed to write to file stream.");
      exit (EXIT_FAILURE);
    }
}

void
check_data (RVL *self)
{
  if (self->data.size <= 0)
    {
      log_fatal ("Size of data is less than 0.");
      exit (EXIT_FAILURE);
    }

  if (self->data.size != rvl_eval_voxels_nbytes (self))
    {
      log_fatal ("Size of data does not match the header information.");
      exit (EXIT_FAILURE);
    }
}

void
check_grid (RVL *self)
{
  if (self->grid.ndx <= 0 || self->grid.ndy <= 0 || self->grid.ndz <= 0)
    {
      log_fatal ("Missing voxel dimensions.");
      exit (EXIT_FAILURE);
    }

  switch (self->grid.type)
    {
    case RVL_GRID_REGULAR:
      if (self->grid.ndx != 1 || self->grid.ndy != 1 || self->grid.ndz != 1)
        {
          log_fatal ("Number of voxel dimensions is not valid "
                         "for regular grid.");
          exit (EXIT_FAILURE);
        }
      break;
    case RVL_GRID_RECTILINEAR:
      {
        u32 *r = self->resolution;
        if (self->grid.ndx != r[0] || self->grid.ndy != r[1]
            || self->grid.ndz != r[2])
          {
            log_fatal ("Number of voxel dimensions do not match "
                           "the resolution.");
            exit (EXIT_FAILURE);
          }
      }
      break;
    default:
      log_fatal ("Invalid grid type: %.2x.", self->grid.type);
      exit (EXIT_FAILURE);
      break;
    }
}
